package AITB_2022.module.complex;

import static rescuecore2.standard.entities.StandardEntityURN.AMBULANCE_TEAM;
import static rescuecore2.standard.entities.StandardEntityURN.CIVILIAN;
import static rescuecore2.standard.entities.StandardEntityURN.FIRE_BRIGADE;
import static rescuecore2.standard.entities.StandardEntityURN.GAS_STATION;
import static rescuecore2.standard.entities.StandardEntityURN.POLICE_FORCE;
import static rescuecore2.standard.entities.StandardEntityURN.REFUGE;
import adf.core.agent.communication.MessageManager;
import adf.core.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.core.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.core.agent.develop.DevelopData;
import adf.core.agent.info.AgentInfo;
import adf.core.agent.info.ScenarioInfo;
import adf.core.agent.info.WorldInfo;
import adf.core.agent.module.ModuleManager;
import adf.core.agent.communication.standard.bundle.centralized.*;
import adf.core.agent.communication.standard.bundle.information.*;
import adf.core.component.communication.CommunicationMessage;
import adf.core.component.module.algorithm.Clustering;
import adf.core.component.module.algorithm.PathPlanning;
import adf.core.component.module.complex.RoadDetector;
import adf.core.debug.DefaultLogger;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.swing.text.html.HTMLDocument.HTMLReader.BlockAction;

import org.apache.log4j.Logger;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;



public class AITRoadDetector extends RoadDetector {
  /**
   * 自身の担当範囲のEntityのEntityIDの集合
   */
  private Set<EntityID> cluster = new HashSet<>();
  /**
   * 完了したタスクの集合
   */
  private Set<EntityID> completed = new HashSet<>();
  /**
   * PFの行動対象となる可能性のあるEntityのEntityIDの集合 key = PFの行動対象となる可能性のあるEntityのEntityID value =
   * keyのタスクの優先度（1（高）〜 8（低））
   */
  private Map<EntityID, Integer> tasks = new HashMap<>();

  private Set<Area> openedAreas = new HashSet<>();
  private Clustering clustering;
  private PathPlanning pathPlanning;

  private EntityID result;
  private Logger logger;

  /**
   * 通信用の定義
   */
  private final boolean VOICE = false;
  private final boolean RADIO = true;
  final private int HELP_BURIED = 5;
  final private int HELP_BLOCKADE = 6;



  public AITRoadDetector(AgentInfo ai, WorldInfo wi, ScenarioInfo si, ModuleManager moduleManager, DevelopData developData) {
    super(ai, wi, si, moduleManager, developData);
    logger = DefaultLogger.getLogger(agentInfo.me());
    this.pathPlanning = moduleManager.getModule(
        "SampleRoadDetector.PathPlanning",
        "adf.impl.module.algorithm.DijkstraPathPlanning");
    this.clustering = moduleManager.getModule("SampleRoadDetector.Clustering",
        "adf.impl.module.algorithm.KMeansClustring");
    registerModule(this.clustering);
    registerModule(this.pathPlanning);
    this.result = null;
  }


  @Override
  public RoadDetector updateInfo(MessageManager messageManager) {
    logger.debug("Time:" + agentInfo.getTime());
    super.updateInfo(messageManager);
    this.receiveMessage(messageManager);
    this.sendMessage(messageManager);
    return this;
  }


  @Override
  public RoadDetector calc() {
    //土木隊の現在位置する建物や道路のIDの取得
    EntityID positionID = this.agentInfo.getPosition();
    StandardEntity currentPosition = worldInfo.getEntity(positionID);
    openedAreas.add((Area) currentPosition);
    //土木隊エージェントが前ステップから移動していないとき
    if (positionID.equals(result)) {
      logger.debug("reach to " + currentPosition + " resetting target");
      sightremoveall(currentPosition);
      this.result = null;
    }

    if (this.result == null) {
        // タスクの優先度がある時の処理
      if (tasks.size() > 0) {
        final Stream<EntityID> candidates =
          this.tasks.keySet()
              .stream()
              .filter(t -> !this.completed.contains(t))
              .sorted(this.comparator1()).limit(5);

        this.result = candidates.min(this.comparator2()).orElse(null);
      }
      
      HashSet<Area> currentTargets = calcTargets();
      logger.debug("Targets: " + currentTargets);
      if (currentTargets.isEmpty()) {
        this.result = null;
        return this;
      }
      this.pathPlanning.setFrom(positionID);
      this.pathPlanning.setDestination(toEntityIds(currentTargets));
      List<EntityID> path = this.pathPlanning.calc().getResult();
      if (path != null && path.size() > 0) {
        this.result = path.get(path.size() - 1);
      }
      logger.debug("Selected Target: " + this.result);
    }
    return this;
  }


  private Collection<EntityID>
      toEntityIds(Collection<? extends StandardEntity> entities) {
    ArrayList<EntityID> eids = new ArrayList<>();
    for (StandardEntity standardEntity : entities) {
      eids.add(standardEntity.getID());
    }
    return eids;
  }


  private HashSet<Area> calcTargets() {
    HashSet<Area> targetAreas = new HashSet<>();
    //ガスステーション？悪さはしてないと思うけど余分
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(REFUGE,
        GAS_STATION)) {
      targetAreas.add((Area) e);
    }
    for (StandardEntity e : this.worldInfo.getEntitiesOfType(CIVILIAN,
        AMBULANCE_TEAM, FIRE_BRIGADE, POLICE_FORCE)) {
      //建物に埋まっている市民，エージェントがいたらターゲットのリストに加える
      //建物ではなく瓦礫の方を探索すべきでは？
      if (isValidHuman(e)) {
        Human h = (Human) e;
        targetAreas.add((Area) worldInfo.getEntity(h.getPosition()));
      }
    }
    HashSet<Area> inClusterTarget = filterInCluster(targetAreas);
    //一回通った場所を再び対象にしないためにターゲットのリストから削除している
    inClusterTarget.removeAll(openedAreas);
    return inClusterTarget;
  }

  //自分のクラスタ内に埋まってる市民orエージェントがいたらそれらをclusterTargetsに格納
  private HashSet<Area> filterInCluster(HashSet<Area> targetAreas) {
    int clusterIndex = clustering.getClusterIndex(this.agentInfo.getID());
    HashSet<Area> clusterTargets = new HashSet<>();
    HashSet<StandardEntity> inCluster = new HashSet<>(
        clustering.getClusterEntities(clusterIndex));
    for (Area target : targetAreas) {
      if (inCluster.contains(target))
        clusterTargets.add(target);

    }
    return clusterTargets;
  }


  @Override
  public EntityID getTarget() {
    return this.result;
  }


  //建物に埋没しているとき
  private boolean isValidHuman(StandardEntity entity) {
    if (entity == null)
      return false;
    if (!(entity instanceof Human))
      return false;

    Human target = (Human) entity;
    //死んでるとき
    if (!target.isHPDefined() || target.getHP() == 0)
      return false;
    if (!target.isPositionDefined())
      return false;
    //ダメージを受けてないとき
    if (!target.isDamageDefined() || target.getDamage() == 0)
      return false;
    if (!target.isBuriednessDefined())
      return false;

    StandardEntity position = worldInfo.getPosition(target);
    if (position == null)
      return false;

    //避難所にいる，または運ばれているとき
    StandardEntityURN positionURN = position.getStandardURN();
    if (positionURN == REFUGE || positionURN == AMBULANCE_TEAM)
      return false;

    return true;
  }

  /**
   * 自身が埋没しているときメッセージを送信して助けを求める
   * @param mm MessageManager
   */
  private void sendMessage(MessageManager mm) {
    final PoliceForce me = (PoliceForce) this.worldInfo.getEntity(this.agentInfo.getID());
    // 自身が埋没しているとき
    if (this.isMyselfBuried()) {
      mm.addMessage(new MessagePoliceForce(RADIO, me, HELP_BURIED, this.agentInfo.getID()));
      mm.addMessage(new MessagePoliceForce(VOICE, me, HELP_BURIED, this.agentInfo.getID()));
      return;
    }
  }

  private boolean isMyselfBuried() {
    Human me = (Human) this.worldInfo.getEntity(this.agentInfo.getID());
    if (!(this.agentInfo.getPositionArea() instanceof Building)) return false;
    if (!me.isBuriednessDefined()) return false;
    return me.getBuriedness() != 0;
  }


  /**
   * エージェントが受け取ったメッセージを整理し，各々のメッセージをthis.handleMessage()で処理をする
   * @param mm MessageManager
   */
  private void receiveMessage(MessageManager mm) {
    //AmbulanceTeamが瓦礫で動けないというメッセージを取得
    final List<CommunicationMessage> helpAmbulance = mm.getReceivedMessageList(MessageAmbulanceTeam.class);
    for(CommunicationMessage mes : helpAmbulance) {
      final MessageAmbulanceTeam message = (MessageAmbulanceTeam) mes;
      this.handleMessage(message);
    }
    //FireBrigadeが瓦礫で動けないというメッセージを取得
    final List<CommunicationMessage> helpFire = mm.getReceivedMessageList(MessageFireBrigade.class);
    for(CommunicationMessage mes : helpFire) {
      final MessageFireBrigade message = (MessageFireBrigade) mes;
      this.handleMessage(message);
    }
  }

  /**
   * MessageFireBrigadeから自身のタスク候補へ追加
   * @param message エージェントが受け取ったMessageFireBrigade
   */
  private void handleMessage(MessageFireBrigade message) {
    final EntityID sender = message.getSenderID();
    final StandardEntity senderEntity = this.worldInfo.getEntity(sender);
    final EntityID targetPosition = message.getTargetID();

    if(message.getAction() != HELP_BLOCKADE) return;
    if(!message.isTargetDefined()) return;

    if(this.worldInfo.getDistance(this.agentInfo.getID(), sender)
            <= this.scenarioInfo.getRawConfig().getIntValue("comms.channels.0.range")) {
      this.putTask(targetPosition, 0);
      this.completed.remove(targetPosition);
    }

    if(!this.cluster.contains(targetPosition)) return;

    this.putTask(targetPosition, 0);
    this.completed.remove(targetPosition);
  }

  /**
   * MessageAmbulanceから自身のタスク候補へ追加
   * @param message エージェントが受け取ったMessageAmbulance
   */
  private void handleMessage(MessageAmbulanceTeam message) {
    final EntityID sender = message.getSenderID();
    final StandardEntity senderEntity = this.worldInfo.getEntity(sender);
    final EntityID targetPosition = message.getTargetID();

    if(message.getAction() != HELP_BLOCKADE) return;
    if(!message.isTargetDefined()) return;

    if(this.worldInfo.getDistance(this.agentInfo.getID(), sender)
            <= this.scenarioInfo.getRawConfig().getIntValue("comms.channels.0.range")) {
      this.putTask(targetPosition, 0);
      this.completed.remove(targetPosition);
    }

    if(!this.cluster.contains(targetPosition)) return;

    this.putTask(targetPosition, 0);
    this.completed.remove(targetPosition);
  }

  /*
   * 瓦礫を撤去して移動する時、視覚内に瓦礫があれば除去してから移動
   * 
   */
  private void sightremoveall(StandardEntity current) {
    // 知覚範囲
    double range = this.scenarioInfo.getPerceptionLosMaxDistance();
    double near = range;
    EntityID taskblock = null;

    // 知覚した瓦礫のインスタンスを抽出する
    Set<Blockade> changedBlockades = this.worldInfo.getChanged().getChangedEntities()
    .stream()
    .map(this.worldInfo::getEntity)
    .filter(Blockade.class::isInstance)
    .map(Blockade.class::cast)
    .collect(Collectors.toSet());

    // 現在地と瓦礫との距離比較
    if (changedBlockades.size() > 0) {
      for(Blockade bl : changedBlockades) {
        double distance = this.worldInfo.getDistance(current, bl);
        if (near > distance) {
          near = distance;
          taskblock = bl.getID();
        }
      }
    }

    // 一番近いものをタスクに追加
    if(taskblock != null) this.putTask(taskblock, 1);
  }

  /**
   * <EntityID, Integer> tasksに登録されているEntityIDを valueとなっている優先度に応じた順序に並び帰るためのComparator
   * valueの値が同じ場合はエージェントからの距離が近い順序になる
   *
   * @return このクラスのフィールドであるtasksに含まれるEntityIDのcomparator
   */
  private Comparator<EntityID> comparator1() {
    final EntityID me = this.agentInfo.getID();


    final Comparator<EntityID> comparator1 = Comparator.comparing(this.tasks::get);
    final Comparator<EntityID> comparator2 = Comparator.comparing(
        i -> this.worldInfo.getDistance(me, i));
    return comparator1.thenComparing(comparator2);
  }

  /**
   * <EntityID, Integer> tasksに登録されているEntityIDを valueとなっている優先度に応じた順序に並び帰るためのComparator
   * valueの値が同じ場合はエージェントのいるAreaからのの距離が近い順序になる
   *
   * @return このクラスのフィールドであるtasksに含まれるEntityIDのcomparator
   */
  private Comparator<EntityID> comparator2() {
    final EntityID position = this.agentInfo.getPosition();

    final Comparator<EntityID> comparator1 = Comparator.comparing(this.tasks::get);
    final Comparator<EntityID> comparator2 = Comparator.comparing(
        i -> this.computeDistance(position, i));
    return comparator1.thenComparing(comparator2);
  }

  /**
   * パスプランニングモジュールを用いて，エージェントが通過するAreaを算出し， 各エリアの中央を順に結んだ線分を用いて，移動距離を算出する．
   *
   * @param from        パスプランニングの開始地点となるAreaのEntityID
   * @param destination パスプランニングの終了地点となるAreaのEntityID
   * @return fromからdestinationを算出された経路に沿って結ぶ線分の長さ（double）
   */
  private double computeDistance(EntityID from, EntityID destination) {
    this.pathPlanning.setFrom(from);
    this.pathPlanning.setDestination(destination);
    this.pathPlanning.calc();
    final List<EntityID> path = this.pathPlanning.getResult();
    final int n = path.size();

    double ret = 0.0;
    for (int i = 1; i < n; ++i) {
      ret += this.worldInfo.getDistance(path.get(i - 1), path.get(i));
    }
    return ret;
  }

  /**
   * PFの行動対象候補となるEntityIDを優先度とともに登録する
   *
   * @param task     PFの行動対象候補であるEntityのEntityID
   * @param priority taskの優先度（1（高）〜8（低））のInteger
   */
  private void putTask(EntityID task, int priority) {
    final int current = this.tasks.getOrDefault(task, Integer.MAX_VALUE);
    this.tasks.put(task, Math.min(current, priority));
  }
}